import logging
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field, PrivateAttr


class Printer:
    def print(self, content: str, color: Optional[str] = None):
        if color == "purple":
            self._print_purple(content)
        elif color == "red":
            self._print_red(content)
        elif color == "green":
            self._print_green(content)
        elif color == "blue":
            self._print_blue(content)
        elif color == "yellow":
            self._print_yellow(content)
        else:
            print(content)

    def _print_purple(self, content):
        print("\033[1m\033[95m {}\033[00m".format(content))

    def _print_green(self, content):
        print("\033[1m\033[92m {}\033[00m".format(content))

    def _print_red(self, content):
        print("\033[91m {}\033[00m".format(content))

    def _print_blue(self, content):
        print("\033[1m\033[94m {}\033[00m".format(content))

    def _print_yellow(self, content):
        print("\033[1m\033[93m {}\033[00m".format(content))


class Logger(BaseModel):
    """
    Control CLI messages.
    Color: red = error, yellow = warning, blue = info (from vhq), green = info (from third parties)
    """

    verbose: bool = Field(default=True)
    info_file_save: bool = Field(default=False, description="whether to save INFO logs")
    _printer: Printer = PrivateAttr(default_factory=Printer)


    def log(self, level: str, message: str, color="yellow"):
        if self.verbose:
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self._printer.print(f"\n{timestamp} - versionHQ [{level.upper()}]: {message}", color=color)

        self._save(level=level, message=message)


    def _save(self, level: str, message: str, filename: str = None):
        import os
        from pathlib import Path

        if level.lower() == "info" and self.info_file_save == False:
            return

        logging_level = logging.INFO
        match level:
            case "warning":
                logging_level = logging.WARNING
            case "error":
                logging_level = logging.ERROR
            case _:
                pass

        cwd = Path.cwd()
        log_file_dir = f"{cwd}/.logs"
        os.makedirs(log_file_dir, exist_ok=True)
        filename = filename if filename else datetime.now().strftime('%H_%M_%S_%d_%m_%Y')
        abs_dir = f"{log_file_dir}/{filename}.log"

        logging.basicConfig(filename=abs_dir, filemode='w', level=logging_level)
        logger = logging.getLogger(__name__)
        file_handler = logging.FileHandler(filename=abs_dir)
        formatter = logging.Formatter('%(asctime)s : %(levelname)s : %(name)s : %(message)s')
        file_handler.setFormatter(formatter)

        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        logger.addHandler(file_handler)

        match logging_level:
            case logging.WARNING:
                logger.warning(message)
            case logging.ERROR:
                logger.error(message)
            case _:
               logger.info(message)



import os
from typing import Any, Dict, cast

from chromadb import Documents, EmbeddingFunction, Embeddings
from chromadb.api.types import validate_embedding_function

from versionhq._utils.logger import Logger


class EmbeddingConfigurator:
    def __init__(self):
        self.embedding_functions = {
            "openai": self._configure_openai,
            "azure": self._configure_azure,
            "ollama": self._configure_ollama,
            "vertexai": self._configure_vertexai,
            "google": self._configure_google,
            "cohere": self._configure_cohere,
            "voyageai": self._configure_voyageai,
            "bedrock": self._configure_bedrock,
            "huggingface": self._configure_huggingface,
            "watson": self._configure_watson,
        }

    def configure_embedder(self, embedder_config: Dict[str, Any] | None = None) -> EmbeddingFunction:
        """
        Configures and returns an embedding function based on the provided config.
        """

        if embedder_config is None:
            return self._create_default_embedding_function()

        provider = embedder_config.get("provider")
        config = embedder_config.get("config", {})
        model_name = config.get("model")

        if isinstance(provider, EmbeddingFunction):
            try:
                validate_embedding_function(provider)
                return provider
            except Exception as e:
                raise ValueError(f"Invalid custom embedding function: {str(e)}")

        if provider not in self.embedding_functions:
            raise Exception(
                f"Unsupported embedding provider: {provider}, supported providers: {list(self.embedding_functions.keys())}"
            )

        return self.embedding_functions[provider](config, model_name)

    @staticmethod
    def _create_default_embedding_function():
        from chromadb.utils.embedding_functions.openai_embedding_function import (
            OpenAIEmbeddingFunction,
        )

        return OpenAIEmbeddingFunction(
            api_key=os.getenv("OPENAI_API_KEY"), model_name="text-embedding-3-small"
        )

    @staticmethod
    def _configure_openai(config, model_name):
        from chromadb.utils.embedding_functions.openai_embedding_function import (
            OpenAIEmbeddingFunction,
        )

        return OpenAIEmbeddingFunction(
            api_key=config.get("api_key") or os.getenv("OPENAI_API_KEY"),
            model_name=model_name,
        )

    @staticmethod
    def _configure_azure(config, model_name):
        from chromadb.utils.embedding_functions.openai_embedding_function import (
            OpenAIEmbeddingFunction,
        )

        return OpenAIEmbeddingFunction(
            api_key=config.get("api_key"),
            api_base=config.get("api_base"),
            api_type=config.get("api_type", "azure"),
            api_version=config.get("api_version"),
            model_name=model_name,
        )

    @staticmethod
    def _configure_ollama(config, model_name):
        from chromadb.utils.embedding_functions.ollama_embedding_function import (
            OllamaEmbeddingFunction,
        )

        return OllamaEmbeddingFunction(
            url=config.get("url", "http://localhost:11434/api/embeddings"),
            model_name=model_name,
        )

    @staticmethod
    def _configure_vertexai(config, model_name):
        from chromadb.utils.embedding_functions.google_embedding_function import (
            GoogleVertexEmbeddingFunction,
        )

        return GoogleVertexEmbeddingFunction(model_name=model_name, api_key=config.get("api_key"))

    @staticmethod
    def _configure_google(config, model_name):
        from chromadb.utils.embedding_functions.google_embedding_function import (
            GoogleGenerativeAiEmbeddingFunction,
        )

        return GoogleGenerativeAiEmbeddingFunction(model_name=model_name, api_key=config.get("api_key"))

    @staticmethod
    def _configure_cohere(config, model_name):
        from chromadb.utils.embedding_functions.cohere_embedding_function import (
            CohereEmbeddingFunction,
        )

        return CohereEmbeddingFunction(
            model_name=model_name,
            api_key=config.get("api_key"),
        )

    @staticmethod
    def _configure_voyageai(config, model_name):
        from chromadb.utils.embedding_functions.voyageai_embedding_function import (
            VoyageAIEmbeddingFunction,
        )

        return VoyageAIEmbeddingFunction(
            model_name=model_name,
            api_key=config.get("api_key"),
        )

    @staticmethod
    def _configure_bedrock(config, model_name):
        from chromadb.utils.embedding_functions.amazon_bedrock_embedding_function import (
            AmazonBedrockEmbeddingFunction,
        )

        return AmazonBedrockEmbeddingFunction(
            session=config.get("session"),
        )

    @staticmethod
    def _configure_huggingface(config, model_name):
        from chromadb.utils.embedding_functions.huggingface_embedding_function import (
            HuggingFaceEmbeddingServer,
        )

        return HuggingFaceEmbeddingServer(
            url=config.get("api_url"),
        )

    @staticmethod
    def _configure_watson(config, model_name):
        try:
            import ibm_watsonx_ai.foundation_models as watson_models
            from ibm_watsonx_ai import Credentials
            from ibm_watsonx_ai.metanames import EmbedTextParamsMetaNames as EmbedParams
        except ImportError as e:
            raise ImportError(
                "IBM Watson dependencies are not installed. Please install them to use Watson embedding."
            ) from e

        class WatsonEmbeddingFunction(EmbeddingFunction):
            def __call__(self, input: Documents) -> Embeddings:
                if isinstance(input, str):
                    input = [input]

                embed_params = {
                    EmbedParams.TRUNCATE_INPUT_TOKENS: 3,
                    EmbedParams.RETURN_OPTIONS: {"input_text": True},
                }

                embedding = watson_models.Embeddings(
                    model_id=config.get("model"),
                    params=embed_params,
                    credentials=Credentials(
                        api_key=config.get("api_key"), url=config.get("api_url")
                    ),
                    project_id=config.get("project_id"),
                )

                try:
                    embeddings = embedding.embed_documents(input)
                    return cast(Embeddings, embeddings)
                except Exception as e:
                    Logger(verbose=True).log(level="error", message=f"Error during Watson embedding: {str(e)}", color="red")
                    raise e

        return WatsonEmbeddingFunction()



import json
import threading
import datetime
import uuid
import inspect
import enum
from textwrap import dedent
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Set, Optional, Callable, Type
from typing_extensions import Annotated, Self

from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator, InstanceOf, field_validator
from pydantic_core import PydanticCustomError

import versionhq as vhq
from versionhq.task.log_handler import TaskOutputStorageHandler
from versionhq.task.evaluate import Evaluation, EvaluationItem
from versionhq.tool.model import Tool, ToolSet
from versionhq._utils import process_config, Logger


class TaskExecutionType(enum.Enum):
    """
    Enumeration to store task execution types of independent tasks without dependencies.
    """
    SYNC = 1
    ASYNC = 2


class ResponseField(BaseModel):
    """
    A class to store a response format that will generate a JSON schema.
    One layer of nested child is acceptable.
    """

    title: str = Field(default=None, description="title of the field")
    data_type: Type = Field(default=None)
    items: Optional[Type] = Field(default=None, description="store data type of the array items")
    properties: Optional[List[BaseModel]] = Field(default=None, description="store dict items in ResponseField format")
    required: bool = Field(default=True)
    nullable: bool = Field(default=False)
    config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="additional rules")


    @model_validator(mode="after")
    def validate_instance(self) -> Self:
        """
        Validate the model instance based on the given `data_type`. (An array must have `items`, dict must have properties.)
        """

        if self.data_type is list and self.items is None:
            self.items = str

        if self.data_type is dict or (self.data_type is list and self.items is dict):
            if self.properties is None:
                raise PydanticCustomError("missing_properties", "The dict type has to set the properties.", {})

            else:
                for item in self.properties:
                    if not isinstance(item, ResponseField):
                        raise PydanticCustomError("invalid_properties", "Properties field must input in ResponseField format.", {})

        return self


    def _format_props(self) -> Dict[str, Any]:
        """
        Structure valid properties from the ResponseField object. 1 layer of nested child is accepted.
        """
        from versionhq.llm.llm_vars import SchemaType

        schema_type = SchemaType(type=self.data_type).convert()
        props: Dict[str, Any] = {}

        if self.data_type is list:
            if self.items is dict:
                nested_p, nested_r = dict(), list()

                if self.properties:
                    for item in self.properties:
                        nested_p.update(**item._format_props())
                        nested_r.append(item.title)

                props = {
                    "type": schema_type,
                    "items": {
                        "type": SchemaType(type=self.items).convert(),
                        "properties": nested_p,
                        "required": nested_r,
                        "additionalProperties": False
                    }
                }

            elif self.items is list:
                props = {
                    "type": schema_type,
                    "items": { "type": SchemaType(type=self.items).convert(), "items": { "type": SchemaType(type=str).convert() }},
                }

            else:
                props = {
                    "type": schema_type,
                    "items": { "type": SchemaType(type=self.items).convert() },
                }


        elif self.data_type is dict:
            p, r = dict(), list()

            if self.properties:
                for item in self.properties:
                    p.update(**item._format_props())
                    r.append(item.title)

            props = {
                "type": schema_type,
                "properties": p,
                "required": r,
                "additionalProperties": False
            }

        else:
            props = {
                "type": schema_type,
                "nullable": self.nullable,
            }

        return { self.title: { **props, **self.config }} if self.config else { self.title: props }


    def _convert(self, value: Any) -> Any:
        """
        Convert the given value to the ideal data type.
        """
        try:
            if self.type is Any:
                pass
            elif self.type is int:
                return int(value)
            elif self.type is float:
                return float(value)
            elif self.type is list or self.type is dict:
                return json.loads(eval(str(value)))
            elif self.type is str:
                return str(value)
            else:
                return value
        except:
            return value


    def _annotate(self, value: Any) -> Annotated:
        """
        Address Pydantic's `create_model`
        """
        return Annotated[self.type, value] if isinstance(value, self.type) else Annotated[str, str(value)]


    def create_pydantic_model(self, result: Dict, base_model: InstanceOf[BaseModel] | Any) -> Any:
        """
        Create a Pydantic model from the given result.
        """
        for k, v in result.items():
            if k is not self.title:
                pass
            elif type(v) is not self.type:
                v = self._convert(v)
                setattr(base_model, k, v)
            else:
                setattr(base_model, k, v)
        return base_model


class TaskOutput(BaseModel):
    """
    A class to store the final output of the given task in raw (string), json_dict, and pydantic class formats.
    """

    task_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store Task ID")
    raw: str = Field(default="", description="Raw output of the task")
    json_dict: Dict[str, Any] = Field(default=None, description="`raw` converted to dictionary")
    pydantic: Optional[Any] = Field(default=None)
    tool_output: Optional[Any] = Field(default=None, description="store tool result when the task takes tool output as its final output")
    callback_output: Optional[Any] = Field(default=None, description="store task or agent callback outcome")
    evaluation: Optional[InstanceOf[Evaluation]] = Field(default=None, description="store overall evaluation of the task output. passed to ltm")


    def to_context_prompt(self) -> str:
        """
        Returns response in string as a prompt context.
        """
        context = ""
        try:
            context = json.dumps(self.json_dict)
        except:
            try:
                if self.pydantic:
                    context = self.pydantic.model_dump()
            except:
                context = self.raw
        return context


    def evaluate(self, task) -> Evaluation:
        """
        Evaluate the output based on the criteria, score each from 0 to 1 scale, and raise suggestions for future improvement.
        """
        from versionhq.task.TEMPLATES.Description import EVALUATE

        self.evaluation = Evaluation() if not self.evaluation else self.evaluation

        # self.evaluation.latency = latency if latency is not None else task.latency
        # self.evaluation.tokens = tokens if tokens is not None else task.tokens

        eval_criteria = task.eval_criteria if task.eval_criteria else  ["Overall competitiveness", ]

        for item in eval_criteria:
            task_eval = Task(
                description=EVALUATE.format(task_description=task.description, task_output=self.raw, eval_criteria=str(item)),
                pydantic_output=EvaluationItem
            )
            res = task_eval.execute(agent=self.evaluation.eval_by)

            if res.pydantic:
                item = EvaluationItem(score=res.pydantic.score, suggestion=res.pydantic.suggestion, criteria=res.pydantic.criteria)
                self.evaluation.items.append(item)

            else:
                try:
                    item = EvaluationItem(
                        score=float(res.json_dict["score"]), suggestion=res.json_dict["suggestion"], criteria=res.json_dict["criteria"]
                    )
                    self.evaluation.items.append(item)
                except Exception as e:
                    Logger(verbose=True).log(level="error", message=f"Failed to convert the evaluation items: {str(e)}", color="red")
                    pass

        return self.evaluation


    @property
    def aggregate_score(self) -> float | int:
        if self.evaluation is None:
            return 0
        else:
            self.evaluation.aggregate_score


    @property
    def json_string(self) -> Optional[str]:
        return json.dumps(self.json_dict)


    def __str__(self) -> str:
        return str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw


class Task(BaseModel):
    """
    A class that stores independent task information and handles task executions.
    """

    __hash__ = object.__hash__
    _original_description: str = PrivateAttr(default=None)
    _task_output_handler = TaskOutputStorageHandler()
    config: Optional[Dict[str, Any]] = Field(default=None, description="values to set on Task class")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="unique identifier for the object, not set by user")
    name: Optional[str] = Field(default=None)
    description: str = Field(description="Description of the actual task")

    # response format
    pydantic_output: Optional[Type[BaseModel]] = Field(default=None, description="store Pydantic class as structured response format")
    response_fields: Optional[List[ResponseField]] = Field(default_factory=list, description="store list of ResponseField as structured response format")

    # tool usage
    tools: Optional[List[ToolSet | Tool | Any]] = Field(default_factory=list, description="tools that the agent can use aside from their tools")
    can_use_agent_tools: bool = Field(default=True, description="whether the agent can use their own tools when executing the task")
    tool_res_as_final: bool = Field(default=False, description="when set True, tools res will be stored in the `TaskOutput`")

    # executing
    execution_type: TaskExecutionType = Field(default=TaskExecutionType.SYNC)
    allow_delegation: bool = Field(default=False, description="whether to delegate the task to another agent")
    callback: Optional[Callable] = Field(default=None, description="callback to be executed after the task is completed.")
    callback_kwargs: Optional[Dict[str, Any]] = Field(default_factory=dict, description="kwargs for the callback when the callback is callable")

    # evaluation
    should_evaluate: bool = Field(default=False, description="True to run the evaluation flow")
    eval_criteria: Optional[List[str]] = Field(default_factory=list, description="criteria to evaluate the outcome. i.e., fit to the brand tone")

    # recording !# REFINEME - eval_callbacks
    processed_agents: Set[str] = Field(default_factory=set, description="store roles of the agents that executed the task")
    tool_errors: int = 0
    delegations: int = 0
    latency: int | float = 0 # job latency in sec
    tokens: int = 0 # tokens consumed
    output: Optional[TaskOutput] = Field(default=None, description="store the final task output in TaskOutput class")


    @model_validator(mode="before")
    @classmethod
    def process_config(cls, values: Dict[str, Any]) -> None:
        return process_config(values_to_update=values, model_class=cls)


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_required_fields(self) -> Self:
        required_fields = ["description",]
        for field in required_fields:
            if getattr(self, field) is None:
                raise ValueError( f"{field} must be provided either directly or through config")
        return self


    @model_validator(mode="after")
    def set_up_tools(self) -> Self:
        if not self.tools:
            pass
        else:
            tool_list = []
            for item in self.tools:
                if isinstance(item, Tool) or isinstance(item, ToolSet):
                    tool_list.append(item)
                elif (isinstance(item, dict) and "function" not in item) or isinstance(item, str):
                    pass
                else:
                    tool_list.append(item) # address custom tool
            self.tools = tool_list
        return self


    def _draft_output_prompt(self, model_provider: str) -> str:
        """
        Draft prompts on the output format by converting `
        """

        output_prompt = ""

        if self.pydantic_output:
            output_prompt = f"""
Your response MUST STRICTLY follow the given repsonse format:
JSON schema: {str(self.pydantic_output)}
"""

        elif self.response_fields:
            output_prompt, output_formats_to_follow = "", dict()
            response_format = str(self._structure_response_format(model_provider=model_provider))
            for item in self.response_fields:
                if item:
                    output_formats_to_follow[item.title] = f"<Return your answer in {item.data_type.__name__}>"

            output_prompt = f"""
Your response MUST be a valid JSON string that strictly follows the response format. Use double quotes for all keys and string values. Do not use single quotes, trailing commas, or any other non-standard JSON syntax.
Response format: {response_format}
Ref. Output image: {output_formats_to_follow}
"""
        else:
            output_prompt = "You MUST Return your response as a valid JSON serializable string, enclosed in double quotes. Do not use single quotes, trailing commas, or other non-standard JSON syntax."

        return dedent(output_prompt)


    def _draft_context_prompt(self, context: Any) -> str:
        """
        Create a context prompt from the given context in any format: a task object, task output object, list, dict.
        """

        context_to_add = None
        if not context:
            # Logger().log(level="error", color="red", message="Missing a context to add to the prompt. We'll return ''.")
            return context_to_add

        match context:
            case str():
                context_to_add = context

            case Task():
                if not context.output:
                    res = context.execute()
                    context_to_add = res.to_context_prompt()

                else:
                    context_to_add = context.output.raw

            case TaskOutput():
                context_to_add = context.to_context_prompt()


            case dict():
                context_to_add = str(context)

            case list():
                res = ", ".join([self._draft_context_prompt(context=item) for item in context])
                context_to_add = res

            case _:
                pass

        return dedent(context_to_add)


    def _prompt(self, model_provider: str = None, context: Optional[Any] = None) -> str:
        """
        Format the task prompt and cascade it to the agent.
        """
        output_prompt = self._draft_output_prompt(model_provider=model_provider)
        task_slices = [self.description, output_prompt, ]

        if context:
            context_prompt = self._draft_context_prompt(context=context)
            task_slices.insert(len(task_slices), f"Consider the following context when responding: {context_prompt}")

        return "\n".join(task_slices)


    def _structure_response_format(self, data_type: str = "object", model_provider: str = "gemini") -> Dict[str, Any] | None:
        """
        Structure a response format either from`response_fields` or `pydantic_output`.
        1 nested item is accepted.
        """

        from versionhq.task.structured_response import StructuredOutput

        response_format: Dict[str, Any] = None

        if model_provider == "openrouter":
            return response_format

        else:
            if self.response_fields:
                properties, required_fields = {}, []
                for i, item in enumerate(self.response_fields):
                    if item:
                        properties.update(item._format_props())
                        required_fields.append(item.title)

                response_schema = {
                    "type": "object",
                    "properties": properties,
                    "required": required_fields,
                    "additionalProperties": False,
                }
                response_format = {
                    "type": "json_schema",
                    "json_schema": { "name": "outcome", "schema": response_schema }
                }


            elif self.pydantic_output:
                response_format = StructuredOutput(response_format=self.pydantic_output, provider=model_provider)._format()

            return response_format


    def _create_json_output(self, raw: str) -> Dict[str, Any]:
        """
        Create json (dict) output from the raw output and `response_fields` information.
        """

        if raw is None or raw == "":
            Logger().log(level="warning", message="The model returned an empty response. Returning an empty dict.", color="yellow")
            output = { "output": "" }
            return output

        try:
            r = str(raw).replace("true", "True").replace("false", "False")
            j = json.dumps(eval(r))
            output = json.loads(j)
            if isinstance(output, dict):
                return output

            else:
                r = str(raw).strip().replace("{'", '{"').replace("{ '", '{"').replace("': '", '": "').replace("'}", '"}').replace("' }", '"}').replace("', '", '", "').replace("['", '["').replace("[ '", '[ "').replace("']", '"]').replace("' ]", '" ]').replace("{\n'", '{"').replace("{\'", '{"').replace("true", "True").replace("false", "False").replace('\"', "'")
                j = json.dumps(eval(r))
                output = json.loads(j)

                if isinstance(output, dict):
                    return output

                else:
                    import ast
                    output = ast.literal_eval(r)
                    return output if isinstance(output, dict) else { "output": str(r) }

        except:
            output = { "output": str(raw) }
            return output


    def _create_pydantic_output(self, raw: str = None, json_dict: Dict[str, Any] = None) -> InstanceOf[BaseModel]:
        """
        Create pydantic output from raw or json_dict output.
        """

        output_pydantic = self.pydantic_output

        try:
            json_dict = json_dict if json_dict else self._create_json_output(raw=raw)

            for k, v in json_dict.items():
                setattr(output_pydantic, k, v)
        except:
            pass

        return output_pydantic


    def interpolate_inputs(self, inputs: Dict[str, Any]) -> None:
        """
        Interpolate inputs into the task description.
        """
        self._original_description = self.description

        if inputs:
            self.description = self._original_description.format(**inputs)


    def _create_short_and_long_term_memories(self, agent: Any, task_output: TaskOutput) -> None:
        """
        After the task execution, create and save short-term/long-term memories in the storage.
        """
        from versionhq.agent.model import Agent
        from versionhq.memory.model import ShortTermMemory, MemoryMetadata, LongTermMemory

        agent = agent if isinstance(agent, Agent) else Agent(role=str(agent), goal=str(agent), with_memory=True)

        if agent.with_memory == False:
            return None

        try:
            evaluation = task_output.evaluation if task_output.evaluation else None
            memory_metadata = evaluation._create_memory_metadata() if evaluation else MemoryMetadata()

            agent.short_term_memory = agent.short_term_memory if agent.short_term_memory else ShortTermMemory(agent=agent, embedder_config=agent.embedder_config)
            agent.short_term_memory.save(
                task_description=str(self.description),
                task_output=str(task_output.raw),
                agent=str(agent.role),
                metadata=memory_metadata
            )

            agent.long_term_memory = agent.long_term_memory if agent.long_term_memory else LongTermMemory()
            agent.long_term_memory.save(
                task_description=str(self.description),
                task_output=str(task_output.raw),
                agent=str(agent.role),
                metadata=memory_metadata
                )

        except AttributeError as e:
            Logger().log(level="error", message=f"Missing attributes for long term memory: {str(e)}", color="red")
            pass

        except Exception as e:
            Logger().log(level="error", message=f"Failed to add to the memory: {str(e)}", color="red")
            pass


    def _build_agent_from_task(self, task_description: str = None) -> InstanceOf["vhq.Agent"]:
        task_description = task_description if task_description else self.description
        if not task_description:
            Logger().log(level="error", message="Task is missing the description.", color="red")
            pass

        agent = vhq.Agent(goal=task_description, role=task_description, maxit=1) #! REFINEME
        return agent


    def _select_agent_to_delegate(self, agent: Any = None) -> Any | None: # return agent object or None
        """
        Creates or selects an agent to delegate the given task and returns Agent object else None.
        """

        from versionhq.agent.model import Agent

        if not self.allow_delegation:
            return None

        agent_to_delegate: InstanceOf[Agent] = None

        if not agent:
            agent_to_delegate = self._build_agent_from_task()

        elif agent and not agent.networks:
            agent_to_delegate = Agent(role="vhq-Delegated-Agent", goal=agent.goal, llm=agent.llm)

        else:
            _managers = []
            _members = []
            for network in agent.networks:
                _managers.extend(member.agent for member in network.members if member.is_manager)
                _members.extend(member.agent for member in network.members if not member.is_manager)

            agent_to_delegate = _managers[0] if _managers else _members[0] if _members else Agent(role="vhq-Delegated-Agent", goal=agent.goal, llm=agent.llm)

        return agent_to_delegate


    # task execution
    def execute(
            self, type: TaskExecutionType = None, agent: Optional["vhq.Agent"] = None, context: Optional[Any] = None
            ) -> TaskOutput | Future[TaskOutput]:
        """
        A main method to handle task execution. Build an agent when the agent is not given.
        """
        type = type if type else  self.execution_type if self.execution_type else TaskExecutionType.SYNC

        if not agent:
            agent = self._build_agent_from_task(task_description=self.description)

        match type:
            case TaskExecutionType.SYNC:
                return self._execute_sync(agent=agent, context=context)

            case TaskExecutionType.ASYNC:
                return self._execute_async(agent=agent, context=context)


    def _execute_sync(self, agent, context: Optional[Any] = None) -> TaskOutput:
        """Executes the task synchronously."""
        return self._execute_core(agent, context)


    def _execute_async(self, agent, context: Optional[Any] = None) -> Future[TaskOutput]:
        """Executes the task asynchronously."""
        future: Future[TaskOutput] = Future()
        threading.Thread(daemon=True, target=self._execute_task_async, args=(agent, context, future)).start()
        return future


    def _execute_task_async(self, agent, context: Optional[str], future: Future[TaskOutput]) -> None:
        """
        Executes the task asynchronously with context handling.
        """
        result = self._execute_core(agent, context)
        future.set_result(result)


    def _execute_core(self, agent, context: Optional[Any]) -> TaskOutput:
        """
        A core method for task execution.
        Handles 1. agent delegation, 2. tools, 3. context to add to the prompt, and 4. callbacks.
        """

        from versionhq.agent.model import Agent
        from versionhq.agent_network.model import AgentNetwork

        task_output: InstanceOf[TaskOutput] = None
        raw_output: str = None
        tool_output: str | list = None
        task_tools: List[List[InstanceOf[Tool]| InstanceOf[ToolSet] | Type[Tool]]] = []
        started_at, ended_at = datetime.datetime.now(), datetime.datetime.now()

        if self.tools:
            for item in self.tools:
                if isinstance(item, ToolSet) or isinstance(item, Tool) or type(item) == Tool:
                    task_tools.append(item)

        if self.allow_delegation == True:
            agent_to_delegate = self._select_agent_to_delegate(agent=agent)
            agent = agent_to_delegate
            self.delegations += 1


        if self.tool_res_as_final == True:
            started_at = datetime.datetime.now()
            tool_output = agent.execute_task(task=self, context=context, task_tools=task_tools)
            ended_at = datetime.datetime.now()
            task_output = TaskOutput(task_id=self.id, tool_output=tool_output, raw=str(tool_output) if tool_output else "")

        else:
            started_at = datetime.datetime.now()
            raw_output = agent.execute_task(task=self, context=context, task_tools=task_tools)
            ended_at = datetime.datetime.now()

            json_dict_output = self._create_json_output(raw=raw_output)
            if "outcome" in json_dict_output:
                json_dict_output = self._create_json_output(raw=str(json_dict_output["outcome"]))

            pydantic_output = self._create_pydantic_output(raw=raw_output, json_dict=json_dict_output) if self.pydantic_output else None

            task_output = TaskOutput(
                task_id=self.id,
                raw=raw_output if raw_output is not None else "",
                pydantic=pydantic_output,
                json_dict=json_dict_output
            )

        self.latency = (ended_at - started_at).total_seconds()
        task_output.evaluation = Evaluation(latency=self.latency, tokens=self.tokens)
        self.output = task_output
        self.processed_agents.add(agent.role)

        if self.should_evaluate and raw_output: # eval only when raw output exsits
            task_output.evaluate(task=self)

        self._create_short_and_long_term_memories(agent=agent, task_output=task_output)

        if self.callback and isinstance(self.callback, Callable):
            kwargs = { **self.callback_kwargs, **task_output.json_dict }
            sig = inspect.signature(self.callback)
            valid_keys = [param.name for param in sig.parameters.values() if param.kind == param.POSITIONAL_OR_KEYWORD]
            valid_kwargs = { k: kwargs[k] if  k in kwargs else None for k in valid_keys }
            callback_res = self.callback(**valid_kwargs)
            task_output.callback_output = callback_res

        # if self.output_file: ## disabled for now
        #     content = (
        #         json_output
        #         if json_output
        #         else pydantic_output.model_dump_json() if pydantic_output else result
        #     )
        #     self._save_file(content)
        return task_output


    def _store_execution_log(self, task_index: int, was_replayed: bool = False, inputs: Optional[Dict[str, Any]] = {}) -> None:
        """
        Store the task execution log.
        """

        self._task_output_handler.update(task=self, task_index=task_index, was_replayed=was_replayed, inputs=inputs)


    @property
    def key(self) -> str:
        output_format = "json" if self.response_fields else "pydantic" if self.pydantic_output is not None else "raw"
        source = [self.description, output_format]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def summary(self) -> str:
        return f"""
Task ID: {str(self.id)}
"Description": {self.description}
"Tools": {", ".join([tool.name for tool in self.tools])}
        """




def test_doc_core_taskoutput_a():
    import versionhq as vhq
    from pydantic import BaseModel

    class CustomOutput(BaseModel):
        test1: str
        test2: list[str]

    def dummy_tool():
        return "dummy"
    def dummy_func(message: str, test1: str, test2: list[str]) -> str:
        return f"""{message}: {test1}, {", ".join(test2)}"""

    task = vhq.Task(
        description="Research a topic to teach a kid aged 6 about math.",
        pydantic_output=CustomOutput,
        tools=[dummy_tool],
        callback=dummy_func,
        callback_kwargs=dict(message="Hi! Here is the result: "),
        should_evaluate=True,
        eval_criteria=["Uniquness", "Fit to audience",],
    )
    res = task.execute()

    assert res.task_id == task.id
    assert res.raw
    assert res.json_dict
    assert res.pydantic.test1 and res.pydantic.test2
    assert "Hi! Here is the result: " in res.callback_output
    assert res.pydantic.test1 in res.callback_output and ", ".join(res.pydantic.test2) in res.callback_output
    assert res.tool_output is None
    assert res.evaluation and isinstance(res.evaluation, vhq.Evaluation)
    assert [isinstance(item, vhq.EvaluationItem) and item.criteria in task.eval_criteria for item in res.evaluation.items]
    assert res.evaluation.latency and res.evaluation.tokens
    assert res.evaluation.aggregate_score is not None and res.evaluation.suggestion_summary



from unittest.mock import patch

from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField, TaskOutput
from versionhq.agent_network.model import AgentNetwork, Member, TaskHandlingProcess, NetworkOutput
from versionhq._utils.usage_metrics import UsageMetrics
from versionhq.llm.model import DEFAULT_MODEL_NAME


def test_form_agent_network():
    agent_a = Agent(role="agent a", goal="My amazing goals", backstory="My amazing backstory", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=True, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
        ],
    )

    assert network.id and network.key and isinstance(network.key, str) and network.managers
    assert task_1 in network.manager_tasks
    assert network.tasks == [task_1, task_2]
    assert [item.id is task_1.id or item.id is task_2.id for item in network.tasks]


def test_form_network_without_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", backstory="My amazing backstory", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME, max_tokens=3000)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ]
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
        ],
    )

    assert network.id and network.key and isinstance(network.key, str)
    assert network.managers is None
    assert network.manager_tasks == []
    assert network.tasks == [task_1, task_2]
    assert [item.id is task_1.id or item.id is task_2.id for item in network.tasks]


def test_launch_without_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[
            ResponseField(title="test1", data_type=str, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    task_2 = Task(
        description="Define the cohort.",
        response_fields=[
            ResponseField(title="test1", data_type=int, required=True),
            ResponseField(title="test2", data_type=list, required=True),
        ],
    )
    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
        ],
    )
    res = network.launch()
    res_all = res.return_all_task_outputs()

    assert isinstance(res, NetworkOutput) and res.network_id is network.id
    assert isinstance(res.raw, str) and isinstance(res.json_dict, dict)
    assert res.pydantic is None
    assert [isinstance(item, TaskOutput) for item in res.task_outputs]
    assert isinstance(res_all, list) and len(res_all) == 2 and [isinstance(item, dict) for item in res_all]
    # assert isinstance(res.token_usage, UsageMetrics)
    # assert res.token_usage.total_tokens == 0 # as we dont set token usage on agent


def test_launch_with_task_callback():
    """
    Each task has callback with callback kwargs.
    """
    demo_list = []
    def demo_callback(item: str) -> None:
        demo_list.append(item)

    agent_a = Agent(
        role="agent a",
        goal="My amazing goals",
        llm=DEFAULT_MODEL_NAME
    )

    agent_b = Agent(
        role="agent b",
        goal="My amazing goals",
        llm=DEFAULT_MODEL_NAME
    )

    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="test1", data_type=str, required=True),],
        callback=demo_callback,
        callback_kwargs=dict(item="pytest demo 1")
    )

    task_2 = Task(
        description="Define the cohort.",
        response_fields=[ResponseField(title="test1", data_type=int, required=True),],
        callback=demo_callback,
        callback_kwargs=dict(item="pytest demo 2")
    )

    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
        ],
    )
    res = network.launch()

    assert res.raw and res.json_dict
    assert len(res.return_all_task_outputs()) == 2
    assert len(demo_list) == 2
    assert "pytest" in demo_list[0]
    assert "pytest" in demo_list[1]


# def test_delegate_in_network():
#     """
#     When the agent belongs to the agent network, its manager or peers are prioritized to delegete the task.
#     """

#     agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
#     agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
#     task_1 = Task(
#         description="Analyze the client's business model.",
#         response_fields=[ResponseField(title="test1", data_type=str, required=True),],
#         allow_delegation=True
#     )
#     task_2 = Task(
#         description="Define the cohort.",
#         response_fields=[ResponseField(title="test1", data_type=int, required=True),],
#         allow_delegation=False
#     )
#     network = AgentNetwork(
#         members=[
#             Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
#             Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
#         ],
#     )
#     res = network.launch()
#
#     assert res.raw is not None
#     assert res.json_dict is not None
#     assert "vhq-Delegated-Agent" in task_1.processed_agents



def test_launch_with_leader():
    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )
    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=True, tasks=[task_2,]),
        ],
    )
    res = network.launch()

    assert isinstance(res, NetworkOutput)
    assert res.network_id is network.id
    assert res.raw is not None
    assert res.json_dict is not None
    assert network.managers[0].agent.id is agent_b.id
    assert len(res.task_outputs) == 2
    assert [item.raw is not None for item in res.task_outputs]
    assert len(network.tasks) == 2
    assert network.tasks[0].output.raw == res.raw


def test_hierarchial_process():
    """
    Manager to handle the top priority task first.
    """

    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_c = Agent(role="agent c", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )
    network = AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=True, tasks=[task_2,]),
            Member(agent=agent_c, is_manager=False)
        ],
        process=TaskHandlingProcess.HIERARCHY
    )
    res = network.launch()

    assert isinstance(res, NetworkOutput)
    assert res.network_id is network.id
    assert res.raw is not None
    assert res.json_dict is not None
    assert network.managers[0].agent.id is agent_b.id
    assert len(res.task_outputs) == 2
    assert [item.raw is not None for item in res.task_outputs]
    assert len(network.tasks) == 2
    assert network.tasks[0].output.raw == res.raw


def test_handle_network_task():
    """
    Make the best network formation with agents and tasks given.
    """

    agent_a = Agent(role="agent a", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_b = Agent(role="agent b", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    agent_c = Agent(role="agent c", goal="My amazing goals", llm=DEFAULT_MODEL_NAME)
    network_task = Task(
        description="Define outbound strategies.",
        response_fields=[ResponseField(title="network_task_1", data_type=str, required=True),],
    )
    task_1 = Task(
        description="Analyze the client's business model.",
        response_fields=[ResponseField(title="task_1", data_type=str, required=True),],
    )
    task_2 = Task(
        description="Define the cohort timeframe.",
        response_fields=[
            ResponseField(title="task_2_1", data_type=int, required=True),
            ResponseField(title="task_2_2", data_type=list, required=True),
        ],
    )

    network_solo = AgentNetwork(
        members=[
            Member(agent=agent_c, is_manager=False)
        ],
        network_tasks=[network_task, task_1, task_2, ]
    )
    network_flat =  AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_c, is_manager=False)
        ],
        network_tasks=[network_task, task_2,]
    )
    network_leader =  AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=True, tasks=[task_2,]),
            Member(agent=agent_c, is_manager=False)
        ],
        network_tasks=[network_task, ]
    )
    network_dual_leaders =  AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=True, tasks=[task_2,]),
            Member(agent=agent_c, is_manager=True)
        ],
        network_tasks=[network_task, ]
    )
    network_leader_without_task =  AgentNetwork(
        members=[
            Member(agent=agent_a, is_manager=False, tasks=[task_1,]),
            Member(agent=agent_b, is_manager=False, tasks=[task_2,]),
            Member(agent=agent_c, is_manager=True)
        ],
        network_tasks=[network_task,]
    )
    networks = [network_solo, network_flat, network_leader, network_dual_leaders, network_leader_without_task]


    for item in networks:
        with patch.object(AgentNetwork, "_execute_tasks", kwargs=dict(tasks=item.tasks)) as private_mock:
            item.launch()
            private_mock.assert_called_once()



from unittest.mock import patch


def test_draft():
    import versionhq as vhq
    from pydantic import BaseModel

    class Trip(BaseModel):
        name: str
        location: str
        description: str
        date: str
        cousine: str
        why_its_suitable: str


    with patch('builtins.input', return_value='Y'):
        task_graph = vhq.workflow(
            Trip,
            context="Planning a suprise day trip for my friend to celebrate her birthday. We live in CA and we like to have Korean food.",
            human=True
        )

        assert task_graph
        assert [k == node.identifier and node.task and isinstance(node, vhq.Node) for k, node in task_graph.nodes.items()]
        assert [isinstance(edge.dependency_type, vhq.DependencyType) and isinstance(edge, vhq.Edge) for k, edge in task_graph.edges.items()]
        assert [k in task_graph.nodes.keys() and v.status == vhq.TaskStatus.NOT_STARTED for k, v in task_graph.nodes.items()]



import litellm

from versionhq.llm.llm_vars import MODELS, LLM_CONTEXT_WINDOW_SIZES
from versionhq.llm.model import LLM, DEFAULT_CONTEXT_WINDOW_SIZE


def dummy_func() -> str:
    return "dummy"


def test_create_llm_from_valid_name():
    """
    Make sure base params are set properly in the LLM class of all the models available in the framework.
    """

    for k, v in MODELS.items():
        for model_name in v:
            llm = LLM(model=model_name, callbacks=[dummy_func,])

            assert llm._init_model_name == model_name
            assert llm.model == model_name
            assert llm.provider == k
            assert llm.context_window_size == int(LLM_CONTEXT_WINDOW_SIZES.get(model_name) *0.75) if LLM_CONTEXT_WINDOW_SIZES.get(model_name) is not None else DEFAULT_CONTEXT_WINDOW_SIZE
            assert llm._supports_function_calling() is not None
            assert llm._supports_stop_words() is not None
            assert litellm.callbacks == [dummy_func,]


def test_create_llm_from_invalid_name():
    """
    Test if all the params will be set properly with a givne invalid model name.
    """
    llm = LLM(model="4o", callbacks=[dummy_func,])

    assert llm._init_model_name == "4o"
    assert llm.model == "gpt-4o"
    assert llm.provider == "openai"
    assert llm.context_window_size == int(128000 * 0.75)
    assert llm._supports_function_calling() == True
    assert llm._supports_stop_words() == True
    assert litellm.callbacks == [dummy_func,]


def test_create_llm_from_provider():
    llm = LLM(provider="gemini", callbacks=[dummy_func,])

    assert llm.model == "gemini/gemini-1.5-flash"
    assert llm.provider == "gemini"
    assert llm.context_window_size == int(LLM_CONTEXT_WINDOW_SIZES.get(llm.model) *0.75)
    assert llm._supports_function_calling() == True
    assert llm._supports_stop_words() == True
    assert litellm.callbacks == [dummy_func,]



from typing import Any, Dict, List


class Storage:
    """Abstract base class defining the storage interface"""

    def save(self, value: Any, metadata: Dict[str, Any]) -> None:
        pass

    def search(self, query: str, limit: int, score_threshold: float) -> Dict[str, Any] | List[Any]:
        return {}

    def reset(self) -> None:
        pass



from typing import List, Optional, Dict, Any
from typing_extensions import Self

from pydantic import BaseModel, Field, model_validator

from versionhq.memory.model import MemoryMetadata

"""
Evaluate task output from accuracy, token consumption, latency perspectives, and mark the score from 0 to 1.
"""


class ScoreFormat:
    def __init__(self, rate: float | int = 0, weight: int = 1):
        self.rate = rate
        self.weight = weight
        self.aggregate = rate * weight


class Score:
    """
    Evaluate the score on 0 (no performance) to 1 scale.
    `rate`: Any float from 0.0 to 1.0 given by an agent.
    `weight`: Importance of each factor to the aggregated score.
    """

    def __init__(
        self,
        brand_tone: ScoreFormat = ScoreFormat(0, 0),
        audience: ScoreFormat = ScoreFormat(0, 0),
        track_record: ScoreFormat = ScoreFormat(0, 0),
        config: Optional[Dict[str, ScoreFormat]] = None
    ):
        self.brand_tone = brand_tone
        self.audience = audience
        self.track_record = track_record
        self.config = config

        if self.config:
            for k, v in self.config.items():
                if isinstance(v, ScoreFormat):
                    setattr(self, k, v)


    def result(self) -> int:
        aggregate_score, denominator = 0, 0

        for k, v in self.__dict__.items():
            aggregate_score += v.aggregate
            denominator += v.weight

        if denominator == 0:
            return 0

        return round(aggregate_score / denominator, 2)


class EvaluationItem(BaseModel):
    """
    A class to store evaluation and suggestion by the given criteria such as accuracy.
    """
    criteria: str
    suggestion: str
    score: float

    def _convert_score_to_score_format(self, weight: int = 1) -> ScoreFormat | None:
        if self.score and isinstance(self.score, float):
            return ScoreFormat(rate=self.score, weight=weight)

        else: return None


class Evaluation(BaseModel):
    items: List[EvaluationItem] = []
    latency: float = Field(default=None, description="job execution latency in seconds")
    tokens: int = Field(default=None, description="tokens consumed")
    eval_by: Any = Field(default=None, description="stores agent object that evaluates the outcome")

    @model_validator(mode="after")
    def set_up_evaluator(self) -> Self:
        from versionhq.agent.inhouse_agents import vhq_task_evaluator
        self.eval_by = vhq_task_evaluator
        return self


    def _create_memory_metadata(self) -> MemoryMetadata:
        """
        Create and store evaluation results in the memory metadata
        """
        eval_by = self.eval_by.role if self.eval_by else None
        score = self.aggregate_score
        eval_criteria = ", ".join([item.criteria for item in self.items]) if self.items else None
        suggestion = self.suggestion_summary
        memory_metadata = MemoryMetadata(eval_by=eval_by, score=score, eval_criteria=eval_criteria, suggestion=suggestion)
        return memory_metadata


    @property
    def aggregate_score(self) -> float:
        """
        Calcurate aggregate score from evaluation items.
        """
        if not self.items:
            return 0

        aggregate_score = 0
        denominator = 0

        for item in self.items:
            score_format = item._convert_score_to_score_format()
            aggregate_score += score_format.aggregate if score_format else 0
            denominator += score_format.weight if score_format else 0

        if denominator == 0:
            return 0

        return round(aggregate_score / denominator, 2)


    @property
    def suggestion_summary(self) -> str | None:
        """
        Return a summary of the suggestions
        """
        if not self.items:
            return None

        summary = ""
        for item in self.items:
            summary += f"{item.suggestion}, "

        return summary



from typing import Any, Dict, Optional, List

from versionhq.memory.model import ShortTermMemory, LongTermMemory, UserMemory


class ContextualMemory:
    """
    A class to construct context from memories (ShortTermMemory, UserMemory).
    The context will be added to the prompt when the agent executes the task.
    """

    def __init__(
        self,
        memory_config: Optional[Dict[str, Any]],
        stm: ShortTermMemory,
        ltm: LongTermMemory,
        um: UserMemory,
        # em: EntityMemory,
    ):
        self.memory_provider = memory_config.get("provider") if memory_config is not None else None
        self.stm = stm
        self.ltm = ltm
        self.um = um


    def build_context_for_task(self, query: str = None) -> str:
        """
        Automatically builds a minimal, highly relevant set of contextual information for a given task.
        """
        if not query:
            return ""

        context = []
        context.append(self._fetch_stm_context(query))
        if self.memory_provider == "mem0":
            context.append(self._fetch_user_context(query))
        return "\n".join(filter(None, context))


    def _fetch_stm_context(self, query) -> str:
        """
        Fetches recent relevant insights from STM related to the task's description and expected_output, formatted as bullet points.
        """
        stm_results = self.stm.search(query)
        formatted_results = "\n".join(
            [
                f"- {result['memory'] if self.memory_provider == 'mem0' else result['context']}"
                for result in stm_results
            ]
        )
        return f"Recent Insights:\n{formatted_results}" if stm_results else ""


    def _fetch_ltm_context(self, task) -> Optional[str]:
        """
        Fetches historical data or insights from LTM that are relevant to the task's description and expected_output, formatted as bullet points.
        """
        ltm_results = self.ltm.search(task, latest_n=2)
        if not ltm_results:
            return None

        formatted_results = [suggestion for result in ltm_results for suggestion in result["metadata"]["suggestions"]]
        formatted_results = list(dict.fromkeys(formatted_results))
        formatted_results = "\n".join([f"- {result}" for result in formatted_results])
        return f"Historical Data:\n{formatted_results}" if ltm_results else ""


    def _fetch_user_context(self, query: str) -> str:
        """
        Fetches and formats relevant user information from User Memory.
        """

        user_memories = self.um.search(query)
        if not user_memories:
            return ""

        formatted_memories = "\n".join(f"- {result['memory']}" for result in user_memories)
        return f"User memories/preferences:\n{formatted_memories}"


    # def _fetch_entity_context(self, query) -> str:
    #     """
    #     Fetches relevant entity information from Entity Memory related to the task's description and expected_output,
    #     formatted as bullet points.
    #     """
    #     em_results = self.em.search(query)
    #     formatted_results = "\n".join(
    #         [
    #             f"- {result['memory'] if self.memory_provider == 'mem0' else result['context']}"
    #             for result in em_results
    #         ]  # type: ignore #  Invalid index type "str" for "str"; expected type "SupportsIndex | slice"
    #     )
    #     return f"Entities:\n{formatted_results}" if em_results else ""



from typing import Optional, Any
from pydantic import InstanceOf

from versionhq.tool.model import ToolSet
from versionhq.tool.cache_handler import CacheHandler, CacheTool


class ToolHandler:
    """
    Record the tool usage by ToolSet instance with cache and error recording.
    """

    last_used_tool: InstanceOf[ToolSet]
    cache: InstanceOf[CacheHandler] = CacheHandler()
    error: Optional[str]
    should_cache: bool

    def __init__(self, last_used_tool: InstanceOf[ToolSet] = None, cache_handler: InstanceOf[CacheHandler] = None, should_cache: bool = True):
        self.last_used_tool = last_used_tool
        self.cache = cache_handler if cache_handler else CacheHandler()
        self.should_cache = should_cache


    def record_last_tool_used(self, last_used_tool: InstanceOf[ToolSet], output: str, should_cache: bool = True) -> None:
        self.last_used_tool = last_used_tool

        if should_cache:
            self.cache = CacheHandler()
            self.cache.add(tool_name=last_used_tool.tool.name, input=str(last_used_tool.kwargs), output=output)


    def has_called_before(self, tool_set: InstanceOf[ToolSet] = None) -> bool:
        if tool_set is None or not self.last_used_tool:
            return False

        if tool_set := self.last_used_tool:
            return bool((tool_set.tool.name == self.last_used_tool.tool.name) and (tool_set.kwargs == self.last_used_tool.kwargs))



from typing import Any, Callable

from pydantic import BaseModel

from versionhq.agent.model import Agent, LLM
from versionhq.task.model import Task, ResponseField
from versionhq.tool.model import Tool

"""
Test a connection to the llm platform (litellm or custom provider's interface).
"""

llms_to_test = [
    "gpt-4o",
    "gemini/gemini-2.0-flash",
    "openrouter/deepseek/deepseek-r1:free",
    "claude-3-haiku",
    "openrouter/google/gemini-2.0-flash-thinking-exp:free",
    "openrouter/google/gemini-2.0-flash-thinking-exp-1219:free",
    "openrouter/google/gemini-2.0-flash-001",
]


class Demo(BaseModel):
    """
    A demo pydantic class to validate the outcome with various nested data types.
    """
    demo_1: int
    demo_2: float
    demo_3: str
    demo_4: bool
    demo_5: list[str]
    demo_6: dict[str, Any]
    demo_nest_1: list[dict[str, Any]]
    demo_nest_2: list[list[str]]
    demo_nest_3: dict[str, list[str]]
    demo_nest_4: dict[str, dict[str, Any]]


demo_response_fields = [
    ResponseField(title="demo_1", data_type=int),
    ResponseField(title="demo_2", data_type=float),
    ResponseField(title="demo_3", data_type=str),
    ResponseField(title="demo_4", data_type=bool),
    ResponseField(title="demo_5", data_type=list, items=str),
    ResponseField(title="demo_6", data_type=dict, properties=[ResponseField(title="demo-item", data_type=str)]),
    ResponseField(title="demo_nest_1", data_type=list, items=str, properties=([
        ResponseField(title="nest1", data_type=dict, properties=[ResponseField(title="nest11", data_type=str)])
    ])), # you can specify field title of nested items
    ResponseField(title="demo_nest_2", data_type=list, items=list),
    ResponseField(title="demo_nest_3", data_type=dict, properties=[
        ResponseField(title="nest1", data_type=list, items=str)
    ]),
    ResponseField(title="demo_nest_4", data_type=dict, properties=[
        ResponseField(title="nest1", data_type=dict, properties=[ResponseField(title="nest12", data_type=str)])
    ])
]


def set_agent(llm: str) -> Agent:
    agent = Agent(role="demo", goal="demo", llm=llm, maxit=1, max_retry_limit=1, max_tokens=3000)
    assert isinstance(agent.llm, LLM)
    assert agent.llm._init_model_name == llm and agent.llm.provider and agent.llm.max_tokens == agent.max_tokens
    return agent


def simple_task(agent: Agent):
    task = Task(description="write a random poem.")
    res = task.execute(agent=agent, context="We are running a test.")

    assert res.raw and res.tool_output is None and res.callback_output is None


def schema_task(agent: Agent):
    task = Task(description="return random values strictly following the given response format.", pydantic_output=Demo)
    res = task.execute(agent=agent, context="We are running a test.")
    assert [
        getattr(res.pydantic, k) and type(getattr(res.pydantic, k)) == v for k, v in Demo.__annotations__.items()
    ]

def res_field_task(agent: Agent):
    task = Task(description="return random values strictly following the given response format.", response_fields=demo_response_fields)
    res = task.execute(agent=agent, context="We are running a test.")
    assert [k in item.title for item in demo_response_fields for k, v in res.json_dict.items()]


def tool_task(agent: Agent):
    class DemoTool(Tool):
        func: Callable[..., Any] = lambda x: "Demo"

    task = Task(
        description="Simply execute the given tools.",
        tools=[DemoTool,],
        tool_res_as_final=True
    )
    res = task.execute(agent=agent, context="We are running a test.")
    assert res.tool_output and res.raw



# comment out
def _test_connection():
    agents = [set_agent(llm=llm) for llm in llms_to_test]
    for agent in agents:
        simple_task(agent=agent)
        tool_task(agent=agent)



"""Test use cases in docs/core/agents.md"""

def test_docs_core_agent_a():
    import versionhq as vhq
    agent = vhq.Agent(
        role="Marketing Analyst",
        goal="Coping with price competition in saturated markets",
    )

    assert agent.id and agent.role == "Marketing Analyst" and agent.goal == "Coping with price competition in saturated markets" and agent.backstory is not None

    res = agent.start(context="Planning a new campaign promotion starting this summer")

    assert isinstance(res, vhq.TaskOutput)
    assert res.json_dict


def test_docs_core_agent_b1():
    from versionhq import Agent
    agent = Agent(
        role="Marketing Analyst",
        goal="Coping with price competition in saturated markets",
        llm="gemini-2.0",
    )

    from versionhq.llm.model import LLM
    assert "gemini-2.0" in agent.llm.model and isinstance(agent.llm, LLM)


# def test_docs_core_agent_b2():
#     import versionhq as vhq

#     agent = vhq.Agent(
#         role="Marketing Analyst",
#         goal="Coping with price competition in saturated markets",
#         llm="gemini-2.0"
#     )

#     agent._update_llm(llm="deepseek", llm_config=dict(max_tokens=3000))
#     assert "deepseek-r1" in agent.llm.model
#     assert agent.llm.max_tokens == 3000


def test_docs_core_agent_b3():
    import versionhq as vhq

    agent = vhq.Agent(
        role="Marketing Analyst",
        goal="Coping with price competition in saturated markets"
    )

    tool = vhq.Tool(func=lambda x: x)
    agent.update(
        tools=[tool], # adding tools
        goal="my new goal", # updating the goal (this will trigger the update of developer_prompt)
        max_rpm=3, # default = 1
        knowledge_sources=["testing", "testing2"], # adding knowledge sources (this will trigger the storage creation)
        memory_config={"user_id": "0000"}, # adding memories
        llm="gemini-2.0", # updating LLM (Valid llm_config will be inherited to the new model.)
        use_developer_prompt=False,
        dummy="I am dummy" # <- invalid field - automatically ignored
    )

    from versionhq.agent.rpm_controller import RPMController
    assert agent.tools == [tool]
    assert agent.goal == "my new goal" and agent.role == "Marketing Analyst" and "my new goal" in agent.backstory
    assert isinstance(agent._rpm_controller, RPMController)
    assert isinstance(agent._knowledge, vhq.Knowledge) and agent.knowledge_sources == ["testing", "testing2"]
    assert isinstance(agent.user_memory, vhq.UserMemory)
    assert "gemini-2.0" in agent.llm.model and agent.llm.provider == "gemini"
    assert agent.use_developer_prompt == False
    assert isinstance(agent.llm, vhq.LLM)


def test_docs_core_agent_c1():
    from versionhq import Agent
    agent = Agent(
        role="Marketing Analyst",
        goal="Coping with price competition in saturated markets"
    )

    assert agent.backstory == "You are an expert marketing analyst with relevant skillsets and abilities to query relevant information from the given knowledge sources. Leveraging these, you will identify competitive solutions to achieve the following goal: coping with price competition in saturated markets."


def test_docs_core_agent_c2():
    from versionhq import Agent

    agent = Agent(
        role="Marketing Analyst",
        goal="Coping with increased price competition in saturated markets.",
        backstory="You are a marketing analyst for a company in a saturated market. The market is becoming increasingly price-competitive, and your company's profit margins are shrinking. Your primary goal is to develop and implement strategies to help your company maintain its market share and profitability in this challenging environment."
    )
    assert agent.backstory == "You are a marketing analyst for a company in a saturated market. The market is becoming increasingly price-competitive, and your company's profit margins are shrinking. Your primary goal is to develop and implement strategies to help your company maintain its market share and profitability in this challenging environment."


def test_docs_core_agent_e1():
    import json
    from typing import Dict, Any
    import versionhq as vhq

    def format_response(res: str = None) -> str | Dict[str, Any]:
        try:
            r = json.dumps(eval(res))
            formatted_res = json.loads(r)
            return formatted_res
        except:
            return res

    agent = vhq.Agent(
        role="Marketing Analyst",
        goal="Coping with increased price competition in saturated markets.",
        callbacks=[format_response]
    )

    import litellm
    assert litellm.callbacks == agent.callbacks


def test_docs_core_agent_e2():
    import json
    from typing import Dict, Any
    import versionhq as vhq


    def assessment(res: str) -> str:
        try:
            sub_agent = vhq.Agent(role="Validator", goal="Validate the given solutions.")
            task = vhq.Task(description=f"""Assess the given solution based on feasibilities and fits to client's strategies, then refine the solution if necessary.
    Solution: {res}
    """)
            r = task.sync_execute(agent=sub_agent)
            return r.raw

        except:
            return res

    def format_response(res: str = None) -> str | Dict[str, Any]:
        try:
            r = json.dumps(eval(res))
            formatted_res = json.loads(r)
            return formatted_res
        except:
            return res

    agent = vhq.Agent(
        role="Marketing Analyst",
        goal="Build solutions to address increased price competition in saturated markets",
        callbacks=[assessment, format_response]
    )

    import litellm
    assert litellm.callbacks == agent.callbacks and len(agent.callbacks) == 2



def test_docs_core_agent_f():
    import versionhq as vhq

    agent = vhq.Agent(
        role="Marketing Analyst",
        goal="Coping with increased price competition in saturated markets.",
        respect_context_window=False,
        max_execution_time=60,
        max_rpm=5,
        llm_config=dict(
                temperature=1,
                top_p=0.1,
                n=1,
                stop="answer",
                dummy="I am dummy" # <- invalid field will be ignored automatically.
            )
        )
    assert isinstance(agent.llm, vhq.LLM)
    assert agent.llm.temperature == 1
    assert agent.llm.top_p == 0.1
    assert agent.llm.n == 1
    assert agent.llm.stop == "answer"


def test_docs_core_agent_g():
    from versionhq import Agent
    from versionhq.task.model import Task
    from versionhq.knowledge.source import StringKnowledgeSource

    content = "Kuriko's favorite color is gold, and she enjoy Japanese food."
    string_source = StringKnowledgeSource(content=content)

    agent = Agent(
        role="Information Agent",
        goal="Provide information based on knowledge sources",
        knowledge_sources=[string_source,]
    )

    task = Task(
        description="Answer the following question: What is Kuriko's favorite color?"
    )

    res = task.execute(agent=agent)
    assert "gold" in res.raw.lower()


def test_docs_core_agent_h():
    from versionhq import Agent

    agent = Agent(
        role="Researcher",
        goal="You research about math.",
        with_memory=True
    )

    from versionhq.memory.model import ShortTermMemory, LongTermMemory
    assert isinstance(agent.short_term_memory, ShortTermMemory) and isinstance(agent.long_term_memory, LongTermMemory)


def test_docs_core_agent_z():
    from versionhq import Agent

    agent = Agent(
        config=dict(
            role="Marketing Analyst",
            goal="Coping with increased price competition in saturated markets.",
        )
    )

    assert agent.role == "Marketing Analyst" and agent.goal == "Coping with increased price competition in saturated markets."



from unittest.mock import patch

def test_minimum_inputs():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task="Describe the color of the sky by weather condition",
        expected_outcome="Dict that pairs color of the sky with specific weather condition"
    )
    assert isinstance(network, vhq.AgentNetwork)
    assert network.members
    assert network.tasks
    assert isinstance(network.formation, vhq.Formation)
    assert network.managers if network.formation in (vhq.Formation.SUPERVISING, vhq.Formation.HYBRID) else not network.managers



def test_specific_formation():
    import versionhq as vhq

    formations_to_test = [
        vhq.Formation.SQUAD,
        vhq.Formation.RANDOM,
        vhq.Formation.SOLO,
        vhq.Formation.SUPERVISING,
        "squad",
        1,
        "dummy",
        1000,
        None
    ]

    for item in formations_to_test:
        network = vhq.form_agent_network(
            task="Describe the color of the sky by weather condition",
            expected_outcome="Dict that pairs color of the sky with specific weather condition",
            formation=item,
        )

        assert isinstance(network, vhq.AgentNetwork)
        assert network.formation == item if isinstance(item, vhq.Formation) else isinstance(network.formation, vhq.Formation)



from typing import List
from versionhq.task.model import Task, TaskOutput


def create_raw_outputs(tasks: List[Task], task_outputs: List[TaskOutput]) -> str:
    """
    Generate string context from the tasks.
    """

    context = ""
    if len(task_outputs) > 0:
        dividers = "\n\n----------\n\n"
        context = dividers.join(output.raw for output in task_outputs)

    else:
        task_outputs_from_task = [
            task.output for task in tasks if task.output is not None
        ]
        dividers = "\n\n----------\n\n"
        context = dividers.join(output.raw for output in task_outputs_from_task)

    return context



import os
import uuid
from abc import ABC
from dotenv import load_dotenv
from typing import Any, Optional, Tuple, Dict
from typing_extensions import Self

from pydantic import BaseModel, Field, model_validator, field_validator, UUID4, PrivateAttr
from pydantic_core import PydanticCustomError

from composio import ComposioToolSet

from versionhq.tool.composio_tool_vars import ComposioAppName, ComposioAuthScheme, composio_app_set, ComposioStatus, ComposioAction
from versionhq.tool.cache_handler import CacheHandler
from versionhq._utils.logger import Logger


load_dotenv(override=True)

DEFAULT_REDIRECT_URL = os.environ.get("DEFAULT_REDIRECT_URL", None)
DEFAULT_USER_ID = os.environ.get("DEFAULT_USER_ID", None)
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)


class ComposioHandler(ABC, BaseModel):
    """
    A class to handle connecting account with Composio and executing actions using Composio ecosystem.
    `connected_account_id` is set up per `app_name` to call the actions on the given app. i.e., salesforce
    """

    _logger: Logger = PrivateAttr(default_factory=lambda: Logger(verbose=True))
    _cache: CacheHandler = PrivateAttr(default_factory=lambda: CacheHandler())

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    app_name: str = Field(default=ComposioAppName.HUBSPOT, max_length=128, description="app name defined by composio")
    user_id: str = Field(default=DEFAULT_USER_ID, description="composio entity id")
    auth_scheme: str = Field(default=ComposioAuthScheme.OAUTH2)
    redirect_url: str = Field(default=DEFAULT_REDIRECT_URL, description="redirect url after successful oauth2 connection")
    connected_account_id: str = Field(
        default=None,
        description="store the client id generated by composio after auth validation. use the id to connect with a given app and execute composio actions"
    )
    tools: Any = Field(default=None, descritpion="retrieved composio tools")

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def validate_app_name(self):
        if self.app_name not in ComposioAppName:
            raise PydanticCustomError("no_app_name", f"The given app name {self.app_name} is not valid.", {})

        return self


    @model_validator(mode="after")
    def validate_auth_scheme(self):
        """
        Raise error when the client uses auth scheme unavailable for the app.
        """
        app_set = next(filter(lambda tup: self.app_name in tup, composio_app_set), None)
        if not app_set:
            raise  PydanticCustomError("no_app_set", f"The app set of {self.app_name} is missing.", {})

        else:
            acceptable_auth_scheme = next(filter(lambda item: self.auth_scheme in item, app_set), None)
            if acceptable_auth_scheme is None:
                raise  PydanticCustomError("invalid_auth_scheme", f"The app {self.app_name} must have different auth_scheme.", {})

        return self


    def _setup_langchain_toolset(self, metadata: Dict[str, Any] = dict()):
        """
        Composio toolset on LangChain for action execution using LLM.
        """
        from composio_langchain import ComposioToolSet
        return ComposioToolSet(api_key=os.environ.get("COMPOSIO_API_KEY"), metadata={**metadata})


    def _connect(
            self, token: Optional[str] = None, api_key: Optional[str] = None, connected_account_id: str = None
        ) -> Tuple[Self | str | Any]:
        """
        Send connection request to Composio, retrieve `connected_account_id`, and proceed with OAuth process of the given app to activate the connection.
        """

        connection_request, connected_account = None, None
        connected_account_id = connected_account_id or self.connected_account_id
        if connected_account_id:
            connected_account = self.toolset.get_connected_account(id=connected_account_id)

            if connected_account and connected_account.status == ComposioStatus.ACTIVE.value:
                return self, ComposioStatus.ACTIVE.value

        if not self.user_id:
            raise PydanticCustomError("entity_id_missing", "Need entity_id to connect with the tool", {})

        if self.auth_scheme == ComposioAuthScheme.API_KEY:
            collected_from_user = {}
            collected_from_user["api_key"] = api_key
            connection_request = self.toolset.initiate_connection(
                connected_account_params = collected_from_user,
                app=self.app_name,
                entity_id=self.user_id,
                auth_scheme=self.auth_scheme,
            )

        if self.auth_scheme == ComposioAuthScheme.BEARER_TOKEN:
            collected_from_user = {}
            collected_from_user["token"] = token
            connection_request = self.toolset.initiate_connection(
                connected_account_params = collected_from_user,
                app=self.app_name,
                entity_id=self.user_id,
                auth_scheme=self.auth_scheme,
            )

        if self.auth_scheme == ComposioAuthScheme.OAUTH2:
            connection_request = self.toolset.initiate_connection(
                app=self.app_name,
                redirect_url = self.redirect_url, # clients will be redirected to this url after successful auth.
                entity_id=self.user_id,
                auth_scheme=self.auth_scheme,
            )

        if connection_request.connectionStatus == ComposioStatus.FAILED.value:
            self._logger.log(level="error", message="Connection to composio failed.", color="red")
            raise PydanticCustomError("connection_failed", "Connection to composio has failed", {})


        connected_account = self.toolset.get_connected_account(id=connection_request.connectedAccountId)
        # Note: connected_account.id === connection_request.connectedAccountId === self.connected_account_id

        if connected_account.status == ComposioStatus.ACTIVE.value:
            setattr(self.toolset, "entity_id", self.user_id)
            self.connected_account_id = connection_request.connectedAccountId

        elif connected_account.status == ComposioStatus.INITIATED.value:
            setattr(self.toolset, "entity_id", self.user_id)
            self.connected_account_id = connection_request.connectedAccountId

            if connection_request.redirectUrl:
                import webbrowser
                webbrowser.open(connection_request.redirectUrl)

        else:
            self._logger.log(level="error", message="The account is invalid.", color="red")
            raise PydanticCustomError("connection_failed", "Connection to composio has failed", {})

        return self, connected_account.status if connected_account else connection_request.connectionStatus


    def execute_composio_action_with_langchain(self, action_name: str | ComposioAction, task_in_natural_language: str) -> Tuple[Self, str]:
        """
        Execute Composio's Action using Langchain's agent ecosystem.
        """
        from langchain import hub
        from langchain_openai import ChatOpenAI
        from langchain.agents import create_openai_functions_agent, AgentExecutor
        from composio_langchain import Action

        action_name = action_name.value if isinstance(action_name, ComposioAction) else action_name
        action = Action(action_name)
        metadata = { action:  { "OPENAI_API_KEY": OPENAI_API_KEY } }
        toolset = self._setup_langchain_toolset(metadata=metadata)
        tools = toolset.get_tools(actions=[action_name,], entity_id=self.user_id)
        if not tools:
            self._logger.log(level="error", message=f"Tools related to {action_name} are not found on Langchain", color="red")
            raise PydanticCustomError("tool_not_found", "Tools not found on Langchain", {})

        self.tools = tools
        llm = ChatOpenAI()
        prompt = hub.pull("hwchase17/openai-functions-agent")
        agent = create_openai_functions_agent(llm, tools, prompt)
        agent_executor = AgentExecutor(agent=agent, tools=tools, verbose=True)
        result = agent_executor.invoke(dict(input=task_in_natural_language))
        return self, result["output"]


    @property
    def toolset(self) -> ComposioToolSet:
        return ComposioToolSet(api_key=os.environ.get("COMPOSIO_API_KEY"))


    def __name__(self):
        return self.app_name



# silence some warnings
import warnings
warnings.filterwarnings(action="ignore", message="Pydantic serializer warnings:", category=UserWarning, module="pydantic.main")
warnings.filterwarnings(action="ignore", category=UserWarning, module="pydantic._internal")
warnings.filterwarnings(action="ignore", module="LiteLLM:utils")

from dotenv import load_dotenv
load_dotenv(override=True)

from versionhq.agent.model import Agent
from versionhq.agent_network.model import AgentNetwork, NetworkOutput, Formation, Member, TaskHandlingProcess
from versionhq.llm.model import LLM
from versionhq.llm.llm_vars import LLM_CONTEXT_WINDOW_SIZES, PARAMS, PROVIDERS, MODELS
from versionhq.clients.customer.model import Customer
from versionhq.clients.product.model import Product, ProductProvider
from versionhq.clients.workflow.model import MessagingWorkflow, MessagingComponent
from versionhq.knowledge.model import Knowledge, KnowledgeStorage
from versionhq.knowledge.source import PDFKnowledgeSource, CSVKnowledgeSource, JSONKnowledgeSource, TextFileKnowledgeSource, ExcelKnowledgeSource, StringKnowledgeSource
from versionhq.knowledge.source_docling import DoclingSource
from versionhq.task_graph.model import TaskStatus, TaskGraph, Node, Edge, DependencyType
from versionhq.task.model import Task, TaskOutput, ResponseField, TaskExecutionType
from versionhq.task.evaluate import Evaluation, EvaluationItem
from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.cache_handler import CacheHandler
from versionhq.tool.tool_handler import ToolHandler
from versionhq.tool.composio_tool import ComposioHandler
from versionhq.memory.contextual_memory import ContextualMemory
from versionhq.memory.model import ShortTermMemory,LongTermMemory, UserMemory, MemoryItem

from versionhq.agent_network.formation import form_agent_network
from versionhq.task_graph.draft import workflow


__version__ = "1.2.1.19"
__all__ = [
    "Agent",

    "AgentNetwork",
    "NetworkOutput",
    "Formation",
    "Member",
    "TaskHandlingProcess",

    "LLM",
    "LLM_CONTEXT_WINDOW_SIZES",
    "PARAMS",
    "PROVIDERS",
    "MODELS",

    "Customer",
    "Product",
    "ProductProvider",
    "MessagingWorkflow",
    "MessagingComponent",

    "Knowledge",
    "KnowledgeStorage",
    "PDFKnowledgeSource",
    "CSVKnowledgeSource",
    "JSONKnowledgeSource",
    "TextFileKnowledgeSource",
    "ExcelKnowledgeSource",
    "StringKnowledgeSource",
    "DoclingSource",

    "TaskStatus",
    "TaskGraph",
    "Node",
    "Edge",
    "DependencyType",

    "Task",
    "TaskOutput",
    "ResponseField",
    "TaskExecutionType",

    "Evaluation",
    "EvaluationItem",

    "Tool",
    "ToolSet",
    "CacheHandler",
    "ToolHandler",
    "ComposioHandler",

    "ContextualMemory",
    "ShortTermMemory",
    "LongTermMemory",
    "UserMemory",
    "MemoryItem",

    "form_agent_network",
    "workflow",
]



import uuid
from abc import ABC
from datetime import datetime
from typing import Any, Dict, List, Optional
from typing_extensions import Self
from pydantic import UUID4, InstanceOf, BaseModel, ConfigDict, Field, field_validator, model_validator
from pydantic_core import PydanticCustomError

from versionhq.clients.product.model import Product
from versionhq.clients.customer.model import Customer
from versionhq.agent.model import Agent
from versionhq.agent_network.model import AgentNetwork
from versionhq.tool.composio_tool_vars import ComposioAppName


class ScoreFormat:
    def __init__(self, rate: float | int = 0, weight: int = 1):
        self.rate = rate
        self.weight = weight
        self.aggregate = rate * weight


class Score:
    """
    Evaluate the score on 0 (no performance) to 1 scale.
    `rate`: Any float from 0.0 to 1.0 given by an agent.
    `weight`: Importance of each factor to the aggregated score.
    """

    def __init__(
        self,
        brand_tone: ScoreFormat = ScoreFormat(0, 0),
        audience: ScoreFormat = ScoreFormat(0, 0),
        track_record: ScoreFormat = ScoreFormat(0, 0),
        **kwargs: Optional[Dict[str, ScoreFormat]],
    ):
        self.brand_tone = brand_tone
        self.audience = audience
        self.track_record = track_record
        self.kwargs = kwargs


    def result(self) -> int:
        aggregate_score = int(self.brand_tone.aggregate) + int(self.audience.aggregate) + int(self.track_record.aggregate)
        denominator = self.brand_tone.weight + self.audience.weight + self.track_record.weight

        for k, v in self.kwargs.items():
            aggregate_score += v.aggregate
            denominator += v.weight

        if denominator == 0:
            return 0

        return round(aggregate_score / denominator, 2)



class MessagingComponent(ABC, BaseModel):
    layer_id: int = Field(default=0, description="add id of the layer: 0, 1, 2")
    message: str = Field(default=None, max_length=1024, description="text message content to be sent")
    score: InstanceOf[Score] = Field(default=None)
    condition: str = Field(default=None, description="condition to execute the next component")
    interval: Optional[str] = Field(default=None, description="ideal interval to set to assess the condition")


    def store_scoring_result(self, subject: str, score_raw: int | Score | ScoreFormat = None) -> Self:
        """
        Set up the `score` field
        """

        if isinstance(score_raw, Score):
            setattr(self, "score", score_raw)

        elif isinstance(score_raw, ScoreFormat):
            score_instance = Score()
            setattr(score_instance, subject, score_raw)
            setattr(self, "score", score_instance)

        elif isinstance(score_raw, int) or isinstance(score_raw, float):
            score_instance, score_format_instance = Score(), ScoreFormat(rate=score_raw, weight=1)
            setattr(score_instance, "kwargs", { subject: score_format_instance })
            setattr(self, "score", score_instance)

        else:
            pass

        return self



class MessagingWorkflow(ABC, BaseModel):
    """
    Store 3 layers of messaging workflow sent to `customer` on the `product`
    """

    _created_at: Optional[datetime]
    _updated_at: Optional[datetime]

    model_config = ConfigDict(extra="allow")

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    messaging_components: List[MessagingComponent] = Field(default_factory=list, description="store messaging components in the workflow")

    # responsible tean or agents
    agent_network: Optional[AgentNetwork] = Field(default=None, description="store a responsibile agent network to autopilot the workflow")
    agents: Optional[List[Agent]] = Field(default=None, description="store responsible agents. None when the `agent_network` fields has a value")

    # metrics
    destination: Optional[ComposioAppName | str] = Field(default=None, description="destination service to launch the workflow")
    product: InstanceOf[Product] = Field(default=None)
    customer: InstanceOf[Customer] = Field(default=None)
    performance_metrics: List[Dict[str, Any]] | List[str] = Field(default=None, max_length=256, description="performance metrics to track")

    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise PydanticCustomError("may_not_set_field", "This field is not to be set by the user.", {})


    @model_validator(mode="after")
    def set_up_destination(self):
        """
        Set up the destination service using ComposioAppName class.
        """
        if isinstance(self.destination, ComposioAppName):
            pass

        elif isinstance(self.destination, str) and self.destination in ComposioAppName:
            self.destination = ComposioAppName(self.destination)

        elif self.destination is None:
            # if self.customer is not None:
            #     self.destination = self.customer.on

            if self.product.provider is not None and self.product.provider.destination_services:
                applied_service = self.product.provider.destination_services[0]
                self.destination = ComposioAppName(applied_service) if applied_service in ComposioAppName else applied_service

        return self


    def reassign_agent(self, agents: List[Agent] = None, agent_network: AgentNetwork = None) -> None:
        """
        Switch agents
        """

        if not agents and not agent_network:
            raise ValueError("Missing agent or agent network to assign.")

        self.agents = agents
        self.agent_network = agent_network
        self.updated_at = datetime.datetime.now()


    @property
    def name(self) -> str:
        if self.customer.id:
            return f"Workflow ID: {self.id} - on {self.product.id} for {self.customer.id}"
        else:
            return f"Workflow ID: {self.id} - on {self.product.id}"



from typing import Dict, Type, List, Any

from pydantic import BaseModel, Field, InstanceOf

from versionhq.llm.llm_vars import SchemaType
from versionhq.llm.model import LLM, DEFAULT_MODEL_PROVIDER_NAME


"""
Generate a JSON schema from the given Pydantic model.
"""


class StructuredObject:
    """
    A class to store the structured dictionary.
    """
    provider: str = None
    field: Type[Field]

    title: str
    dtype: str = "object"
    properties: Dict[str, Dict[str, str]] = dict()
    required: List[str] = list()
    additionalProperties: bool = False

    def __init__(self, name, field: Type[Field], provider: str | InstanceOf[LLM] = None):
        self.title = name
        self.field = field
        self.dtype = "object"
        self.additionalProperties = False
        self.provider = provider if isinstance(provider, str) else provider.provider if isinstance(provider, LLM) else DEFAULT_MODEL_PROVIDER_NAME

    def _format(self):
        if not self.field:
            pass
        else:
            description = self.field.description if hasattr(self.field, "description") and self.field.description is not None else ""
            field_name = self.field.__name__ if hasattr(self.field, "__name__") and self.field.__name__ else self.title
            self.properties.update({ field_name : { "type": SchemaType(self.field.annotation.__args__).convert() }})
            self.required.append(field_name)

            return {
                self.title: {
                    "type": self.dtype,
                    "description": description,
                    "properties": self.properties,
                    "additionalProperties": self.additionalProperties,
                    "required": self.required
                }
            }



class StructuredList:
    """
    A class to store a structured list with 1 nested object.
    """
    provider: str = DEFAULT_MODEL_PROVIDER_NAME
    field: Type[Field]
    title: str = ""
    dtype: str = "array"
    items: Dict[str, Dict[str, str]] = dict()

    def __init__(self, name, field: Type[Field], provider: str | LLM = DEFAULT_MODEL_PROVIDER_NAME):
        self.provider = provider if isinstance(provider, str) else provider.provider
        self.field = field
        self.title = name
        self.dtype = "array"
        self.items = dict()


    def _format(self):
        field = self.field
        if not field:
            pass
        else:
            description = "" if field.description is None else field.description
            props = {}

            for i, item in enumerate(field.annotation.__args__):
                nested_object_type = item.__origin__ if hasattr(item, "__origin__") else item

                if nested_object_type == dict:
                    props.update({
                        # "nest":  {
                            "type": "object",
                            "properties": { f"{str(i)}": { "type": "string"} },
                            "required": [f"{str(i)}",],
                            "additionalProperties": False
                        # }
                        })

                elif nested_object_type == list:
                    props.update({
                        # "nest":  {
                            "type": "array",
                            "items": { "type": "string" },
                        # }
                        })
                else:
                    props.update({ "type": SchemaType(nested_object_type).convert() })

            self.items = { **props }

            return {
                 self.title: {
                    "type": self.dtype,
                    "description": description,
                    "items": self.items,
                }
            }


class StructuredOutput(BaseModel):
    response_format: Any = None # pydantic base model
    provider: str = None
    applicable_models: List[InstanceOf[LLM] | str] = list()
    name: str = ""
    schema: Dict[str, Any] = dict(type="object", additionalProperties=False, properties=dict(), required=list())


    def _format(self, **kwargs):
        if self.response_format is None:
            pass

        self.name = self.response_format.__name__

        for name, field in self.response_format.model_fields.items():
            self.schema["required"].append(name)

            if hasattr(field.annotation, "__origin__") and field.annotation.__origin__ == dict:
                self.schema["properties"].update(StructuredObject(name=name, field=field)._format())

            elif hasattr(field.annotation, "__origin__") and field.annotation.__origin__ == list:
                self.schema["properties"].update(StructuredList(name=name, field=field)._format())
            else:
                self.schema["properties"].update({ name: { "type": SchemaType(field.annotation).convert(), **kwargs }})

        return {
            "type": "json_schema",
            "json_schema": { "name": self.name, "schema": self.schema }
        }



"""frontend home use cases pre-test"""

def test_demo_hello_world():
    import versionhq as vhq

    task = vhq.Task(description='hello world')
    res = task.execute()

    assert res.raw


def test_demo_network():
    import versionhq as vhq

    network = vhq.form_agent_network(
        task='draft a promotional email for the given client',
        expected_outcome='email subject and body in string',
        context="use email_subject and email_body as keys in your response."
    )

    assert isinstance(network, vhq.AgentNetwork) and isinstance(network.formation, vhq.Formation)

    res = network.launch()

    assert res.pydantic is not None



def test_demo_agent_customization():
    import versionhq as vhq
    import pathlib
    current_path = pathlib.Path(__file__).parent.resolve()

    agent = vhq.Agent(
        role='Demo Manager',
        goal='run a demo successfully'
    )

    agent.update(
        llm='gemini-2.0',
        llm_config = dict(
            temperature=1,
            top_p=0.1,
            n=1,
            stop="test",
        ),
        knowledge_sources = [
            'https://business.linkedin.com',
            f'{current_path}/demo.csv',
        ],
        with_memory=True
    )

    assert "gemini-2.0" in agent.llm.model
    assert agent.llm.temperature == 1 and agent.llm.top_p == 0.1 and agent.llm.n==1 and agent.llm.stop == "test"
    assert agent.knowledge_sources == ['https://business.linkedin.com', f'{current_path}/demo.csv',]
    assert agent.with_memory == True



import uuid
import warnings
from enum import Enum
from concurrent.futures import Future
from hashlib import md5
from typing import Any, Dict, List, Callable, Optional, Tuple
from pydantic import UUID4, BaseModel, Field, PrivateAttr, field_validator, model_validator
from pydantic._internal._generate_schema import GenerateSchema
from pydantic_core import PydanticCustomError, core_schema

from versionhq.agent.model import Agent
from versionhq.task.model import Task, TaskOutput, TaskExecutionType, ResponseField
from versionhq.task.formatter import create_raw_outputs
from versionhq._utils.logger import Logger
from versionhq._utils.usage_metrics import UsageMetrics


initial_match_type = GenerateSchema.match_type

def match_type(self, obj):
    if getattr(obj, "__name__", None) == "datetime":
        return core_schema.datetime_schema()
    return initial_match_type(self, obj)


GenerateSchema.match_type = match_type
warnings.filterwarnings("ignore", category=SyntaxWarning, module="pysbd")


class Formation(str, Enum):
    SOLO = 1
    SUPERVISING = 2
    SQUAD = 3
    RANDOM = 4
    HYBRID = 10


class TaskHandlingProcess(str, Enum):
    """
    A class representing task handling processes to tackle multiple tasks.
    When the agent network has multiple tasks that connect with edges, follow the edge conditions.
    """
    SEQUENT = 1
    HIERARCHY = 2
    CONSENSUAL = 3 # either from managers or peers or (human) - most likely controlled by edge


class NetworkOutput(TaskOutput):
    """
    A class to store output from the network, inherited from TaskOutput class.
    """

    network_id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True, description="store the network ID as an identifier that generate the output")
    task_description: str = Field(default=None, description="store initial request (task description) from the client")
    task_outputs: list[TaskOutput] = Field(default=list, description="store TaskOutput objects of all tasks that the network has executed")
    # token_usage: UsageMetrics = Field(default=dict, description="processed token summary")

    def return_all_task_outputs(self) -> List[Dict[str, Any]]:
        res = [output.json_dict for output in self.task_outputs]
        return res

    def __str__(self):
        return (str(self.pydantic) if self.pydantic else str(self.json_dict) if self.json_dict else self.raw)

    def __getitem__(self, key):
        if self.pydantic and hasattr(self.pydantic, key):
            return getattr(self.pydantic, key)
        elif self.json_dict and key in self.json_dict:
            return self.json_dict[key]
        else:
            raise KeyError(f"Key '{key}' not found in the output.")



class Member(BaseModel):
    """
    A class to store a member (agent) in the network, with its tasks and memory/knowledge share settings.
    """
    agent: Agent | None = Field(default=None)
    is_manager: bool = Field(default=False)
    can_share_knowledge: bool = Field(default=True, description="whether to share the agent's knowledge in the network")
    can_share_memory: bool = Field(default=True, description="whether to share the agent's memory in the network")
    tasks: Optional[List[Task]] = Field(default_factory=list, description="tasks explicitly assigned to the agent")

    @property
    def is_idling(self):
        return bool(self.tasks)


class AgentNetwork(BaseModel):
    """
    A class to store a agent network with agent members and their tasks.
    Tasks can be 1. multiple individual tasks, 2. multiple dependant tasks connected via TaskGraph, and 3. hybrid.
    """

    __hash__ = object.__hash__
    _execution_span: Any = PrivateAttr()
    _inputs: Optional[Dict[str, Any]] = PrivateAttr(default=None)

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: Optional[str] = Field(default=None)
    members: List[Member] = Field(default_factory=list)
    formation: Optional[Formation] = Field(default=None)
    should_reform: bool = Field(default=False, description="True if task exe. failed or eval scores below threshold")

    # formation planning
    network_tasks: Optional[List[Task]] = Field(default_factory=list, description="tasks without dedicated agents")

    # task execution rules
    prompt_file: str = Field(default="", description="absolute path to the prompt json file")
    process: TaskHandlingProcess = Field(default=TaskHandlingProcess.SEQUENT)

    # callbacks
    pre_launch_callbacks: List[Callable[[Optional[Dict[str, Any]]], Optional[Dict[str, Any]]]] = Field(
        default_factory=list,
        description="list of callback functions to be executed before the network launch. i.e., adjust inputs"
    )
    post_launch_callbacks: List[Callable[[NetworkOutput], NetworkOutput]] = Field(
        default_factory=list,
        description="list of callback functions to be executed after the network launch. i.e., store the result in repo"
    )
    step_callback: Optional[Any] = Field(default=None, description="callback to be executed after each step for all agents execution")

    cache: bool = Field(default=True)
    execution_logs: List[Dict[str, Any]] = Field(default_factory=list, description="list of execution logs of the tasks handled by members")
    usage_metrics: Optional[UsageMetrics] = Field(default=None, description="usage metrics for all the llm executions")


    def __name__(self) -> str:
        return self.name if self.name is not None else self.id.__str__


    @field_validator("id", mode="before")
    @classmethod
    def _deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        """Prevent manual setting of the 'id' field by users."""
        if v:
            raise PydanticCustomError("may_not_set_field", "The 'id' field cannot be set by the user.", {})


    @model_validator(mode="after")
    def assess_tasks(self):
        """
        Validates if the model recognize all tasks that the network needs to handle.
        """

        if self.tasks:
            if all(task in self.tasks for task in self.network_tasks) == False:
                raise PydanticCustomError("task_validation_error", "`network_tasks` needs to be recognized in the task.", {})


            num_member_tasks = 0
            for member in self.members:
                num_member_tasks += len(member.tasks)

            # if len(self.tasks) != len(self.network_tasks) + num_member_tasks:
            #     raise PydanticCustomError("task_validation_error", "Some tasks are missing.", {})
        return self


    @model_validator(mode="after")
    def check_manager(self):
        """
        Check if the agent network has a manager
        """
        if self.process == TaskHandlingProcess.HIERARCHY or self.formation == Formation.SUPERVISING:
            if not self.managers:
                Logger().log(level="error", message="The process or formation created needs at least 1 manager agent.", color="red")
                raise PydanticCustomError("missing_manager", "`manager` is required when using hierarchical process.", {})

        ## comment out for the formation flexibilities
        # if self.managers and (self.manager_tasks is None or self.network_tasks is None):
        #     Logger().log(level="error", message="The manager is idling. At least 1 task needs to be assigned to the manager.", color="red")
        #     raise PydanticCustomError("missing_manager_task", "manager needs to have at least one manager task or network task.", {})

        return self


    @model_validator(mode="after")
    def validate_task_member_paring(self):
        """
        Sequential task processing without any network_tasks require a task-agent pairing.
        """
        if self.process == TaskHandlingProcess.SEQUENT and self.network_tasks is None:
            for task in self.tasks:
                if not [member.task == task for member in self.members]:
                    Logger().log(level="error", message=f"The following task needs a dedicated agent to be assinged: {task.description}", color="red")
                    raise PydanticCustomError("missing_agent_in_task", "Sequential process error: Agent is missing the task", {})
        return self


    @model_validator(mode="after")
    def validate_end_with_at_most_one_async_task(self):
        """
        Validates that the agent network completes max. one asynchronous task by counting tasks traversed backward
        """

        async_task_count = 0
        for task in reversed(self.tasks):
            if not task:
                break
            elif task.execution_type == TaskExecutionType.ASYNC:
                async_task_count += 1
            else:
                break

        if async_task_count > 1:
            raise PydanticCustomError("async_task_count", "The agent network must end with at maximum one asynchronous task.", {})
        return self


    @staticmethod
    def handle_assigning_agents(unassigned_tasks: List[Task]) -> List[Member]:
        """
        Build an agent and assign it with a task. Return a list of Member connecting the agent created and the task given.
        """

        from versionhq.agent.inhouse_agents import vhq_agent_creator

        new_member_list: List[Member] = []

        for unassgined_task in unassigned_tasks:
            task = Task(
                description=f"""
                    Based on the following task summary, draft a AI agent's role and goal in concise manner.
                    Task summary: {unassgined_task.summary}
                """,
                response_fields=[
                    ResponseField(title="goal", data_type=str, required=True),
                    ResponseField(title="role", data_type=str, required=True),
                ],
            )
            res = task.execute(agent=vhq_agent_creator)
            agent = Agent(
                role=res.json_dict["role"] if "role" in res.json_dict else res.raw,
                goal=res.json_dict["goal"] if "goal" in res.json_dict else task.description
            )
            if agent.id:
                member = Member(agent=agent, tasks=[unassgined_task], is_manager=False)
                new_member_list.append(member)

        return new_member_list


    def _get_responsible_agent(self, task: Task) -> Agent | None:
        if task is None:
            return None
        else:
            for member in self.members:
                if member.tasks and [item for item in member.tasks if item.id == task.id]:
                    return member.agent

            return None


    def _assign_tasks(self) -> None:
        """
        Form a agent network considering given agents and tasks, and update `self.members` field:
            1. Idling managers to take the network tasks.
            2. Idling members to take the remaining tasks starting from the network tasks to member tasks.
            3. Create agents to handle the rest tasks.
        """

        idling_managers: List[Member] = [member for member in self.members if member.is_idling and member.is_manager == True]
        idling_members: List[Member] =  [member for member in self.members if member.is_idling and member.is_manager == False]
        unassigned_tasks: List[Task] = self.network_tasks + self.member_tasks_without_agent if self.network_tasks else self.member_tasks_without_agent
        new_members: List[Member] = []

        if idling_managers:
            idling_managers[0].tasks.extend(unassigned_tasks)

        elif idling_members:
            idling_members[0].tasks.extend(unassigned_tasks)

        else:
            new_members = self.handle_assigning_agents(unassigned_tasks=unassigned_tasks)
            if new_members:
                self.members += new_members


    # task execution
    def _process_async_tasks(self, futures: List[Tuple[Task, Future[TaskOutput], int]], was_replayed: bool = False) -> List[TaskOutput]:
        """
        When we have `Future` tasks, updated task outputs and task execution logs accordingly.
        """

        task_outputs: List[TaskOutput] = []

        for future_task, future, task_index in futures:
            task_output = future.result()
            task_outputs.append(task_output)
            future_task._store_execution_log(task_index, was_replayed)

        return task_outputs


    def _calculate_usage_metrics(self) -> UsageMetrics:
        """
        Calculate and return the usage metrics that consumed by the agent network.
        """
        total_usage_metrics = UsageMetrics()

        for member in self.members:
            agent = member.agent
            if hasattr(agent, "_token_process"):
                token_sum = agent._token_process.get_summary()
                total_usage_metrics.add_usage_metrics(token_sum)

        if self.managers:
            for manager in self.managers:
                if hasattr(manager.agent, "_token_process"):
                    token_sum = manager.agent._token_process.get_summary()
                    total_usage_metrics.add_usage_metrics(token_sum)

        self.usage_metrics = total_usage_metrics
        return total_usage_metrics


    def _execute_tasks(self, tasks: List[Task], start_index: Optional[int] = 0, was_replayed: bool = False) -> NetworkOutput:
        """
        Executes tasks sequentially and returns the final output in NetworkOutput class.
        When we have a manager agent, we will start from executing manager agent's tasks.
        Priority:
        1. Network tasks > 2. Manager task > 3. Member tasks (in order of index)
        """

        task_outputs: List[TaskOutput] = []
        lead_task_output: TaskOutput = None
        futures: List[Tuple[Task, Future[TaskOutput], int]] = []
        last_sync_output: Optional[TaskOutput] = None

        for task_index, task in enumerate(tasks):
            if start_index is not None and task_index < start_index:
                if task.output:
                    if task.execution_type == TaskExecutionType.ASYNC:
                        task_outputs.append(task.output)
                    else:
                        task_outputs = [task.output]
                        last_sync_output = task.output
                continue

            responsible_agent = self._get_responsible_agent(task)
            if responsible_agent is None:
                self._assign_tasks()

            ## commented out - this will be handled by node objects
            # if isinstance(task, ConditionalTask):
            #     skipped_task_output = task._handle_conditional_task(task_outputs, futures, task_index, was_replayed)
            #     if skipped_task_output:
            #         continue

            # self._log_task_start(task, responsible_agent)

            if task.execution_type == TaskExecutionType.ASYNC:
                context = create_raw_outputs(tasks=[task, ], task_outputs=([last_sync_output,] if last_sync_output else []))
                future = task._execute_async(agent=responsible_agent, context=context)
                futures.append((task, future, task_index))
            else:
                context = create_raw_outputs(tasks=[task,], task_outputs=([last_sync_output,] if last_sync_output else [] ))
                task_output = task.execute(agent=responsible_agent, context=context)

                if self.managers and responsible_agent in [manager.agent for manager in self.managers]:
                    lead_task_output = task_output

                task_outputs.append(task_output)
                task._store_execution_log(task_index, was_replayed, self._inputs)


        if futures:
            task_outputs = self._process_async_tasks(futures, was_replayed)

        if not task_outputs:
            Logger().log(level="error", message="Missing task outputs.", color="red")
            raise ValueError("Missing task outputs")

        final_task_output = lead_task_output if lead_task_output is not None else task_outputs[0] #! REFINEME

        # token_usage = self._calculate_usage_metrics() #! combining with Eval

        return NetworkOutput(
            network_id=self.id,
            raw=final_task_output.raw,
            json_dict=final_task_output.json_dict,
            pydantic=final_task_output.pydantic,
            task_outputs=task_outputs,
            # token_usage=token_usage,
        )


    def launch(self, kwargs_pre: Optional[Dict[str, str]] = None, kwargs_post: Optional[Dict[str, Any]] = None) -> NetworkOutput:
        """
        Launch the agent network - executing tasks and recording their outputs.
        """

        metrics: List[UsageMetrics] = []

        if self.network_tasks or self.member_tasks_without_agent:
            self._assign_tasks()

        if kwargs_pre is not None:
            for func in self.pre_launch_callbacks: # signature check
                func(**kwargs_pre)

        for member in self.members:
            agent = member.agent
            agent.networks.append(self)

            if self.step_callback:
                agent.callbacks.append(self.step_callback)

        if self.process is None:
            self.process = TaskHandlingProcess.SEQUENT

        result = self._execute_tasks(self.tasks)

        for func in self.post_launch_callbacks:
            result = func(result, **kwargs_post)

        metrics += [member.agent._token_process.get_summary() for member in self.members]

        self.usage_metrics = UsageMetrics()
        for metric in metrics:
            self.usage_metrics.add_usage_metrics(metric)

        return result


    @property
    def key(self) -> str:
        source = [str(member.agent.id.__str__) for member in self.members] + [str(task.id.__str__) for task in self.tasks]
        return md5("|".join(source).encode(), usedforsecurity=False).hexdigest()


    @property
    def managers(self) -> List[Member] | None:
        managers = [member for member in self.members if member.is_manager == True]
        return managers if len(managers) > 0 else None


    @property
    def manager_tasks(self) -> List[Task]:
        """
        Tasks (incl. network tasks) handled by managers in the agent network.
        """
        res = list()

        if self.managers:
            for manager in self.managers:
                if manager.tasks:
                    res.extend(manager.tasks)

        return res


    @property
    def tasks(self) -> List[Task]:
        """
        Return all the tasks that the agent network needs to handle in order of priority:
        1. network_tasks, -> assigned to the member
        2. manager_task,
        3. members' tasks
        """

        network_tasks = self.network_tasks
        manager_tasks = self.manager_tasks
        member_tasks = []

        for member in self.members:
            if member.is_manager == False and member.tasks:
                a = [item for item in member.tasks if item not in network_tasks and item not in manager_tasks]
                member_tasks += a

        return network_tasks + manager_tasks + member_tasks


    @property
    def member_tasks_without_agent(self) -> List[Task]:
        res = list()

        if self.members:
            for member in self.members:
                if member.agent is None and member.tasks:
                    res.extend(member.tasks)

        return res
